import json
from datetime import datetime

from flask_scenario_testing.analysis.Scenario import Scenario
from flask_scenario_testing.analysis.support.Measurement import Measurement


class Results(object):

    def __init__(self, results):
        self._results = results

    @staticmethod
    def from_json(absolute_file_name: str):
        with open(absolute_file_name, 'r') as f:
            configuration = json.load(f)

            return Results(configuration)

    def get_total_request_count(self):
        total_count = 0

        for endpoint in self.endpoints():
            endpoint_request_count = len(endpoint['latency_measurements'])

            total_count += endpoint_request_count

        return total_count

    def endpoints(self):
        return self._results['endpoints']

    def started_at(self):
        return self._results['started_at']

    def stopped_at(self):
        return self._results['stopped_at']

    def scenarios(self):
        return [
            Scenario(
                name=r['name'],
                started_at=datetime.fromisoformat(r['start']),
                ended_at=datetime.fromisoformat(r['end'])
            ) for idx, r in enumerate(self._results['scenarios'])
        ]

    def cpu_usage_measurements(self) -> [Measurement]:
        results = self._results['cpu_usage_measurements']

        return [Measurement(value=c['measurement'], time=datetime.fromisoformat(c['time'])) for c in results]

    def _get_endpoint(self, endpoint_name):
        filtered = list(filter(lambda x: x['name'] == endpoint_name, self.endpoints()))

        if len(filtered) == 0:
            return None

        return filtered[0]

    def latency_measurements(self, endpoint_name: str, start: datetime = None, end: datetime = None) -> [Measurement]:
        endpoint = self._get_endpoint(endpoint_name)

        if not endpoint:
            raise ValueError('Endpoint not found')

        if start and end:
            return [
                Measurement(time=measurement_time, value=m['measurement'])
                for m in endpoint['latency_measurements']
                if start <= (measurement_time := datetime.fromisoformat(m['time'])) <= end
            ]
        else:
            return [
                Measurement(time=datetime.fromisoformat(m['time']), value=m['measurement'])
                for m in endpoint['latency_measurements']
            ]


from flask_scenario_testing.analysis.Results import Results
from flask_scenario_testing.analysis.report_sections.ReportSection import ReportSection


class SimulationDetails(ReportSection):
    def print(self, results: Results):
        self.table(
            headers=['', ''],
            data=[
                ['Total number of requests', results.get_total_request_count()],
                ['Simulation start', results.started_at()],
                ['Simulation start', results.stopped_at()],
            ]
        )

        rows = []

        for idx, scenario in enumerate(results.scenarios()):
            rows.append([str(idx + 1),  'Start',    scenario.started_at()])
            rows.append(['',            'End',      scenario.ended_at()])

        self.table(['Scemarop', '', ''], rows)



from flask_scenario_testing.analysis.Results import Results
from flask_scenario_testing.analysis.report_sections.ReportSection import ReportSection


class EndpointOverview(ReportSection):
    def print(self, results: Results):
        endpoints = results.endpoints()

        self.heading('Endpoint overview', 3)

        self.table(
            headers=['Name', 'Requests measured'],
            data=[
                [endpoint['name'], len(endpoint['latency_measurements'])] for endpoint in endpoints
            ]
        )



from datetime import datetime
from typing import Optional
import random

class RunningScenario(object):

    def __init__(self, name: str, started_at: datetime, modifiers: dict, options: dict):
        self._name = name
        self._started_at = started_at
        self._modifiers = modifiers
        self._stopped_at = None
        self._options = options
        random.seed(0)

    def name(self) -> str:
        return self._name

    def options(self) -> dict:
        return self._options

    def started_at(self) -> datetime:
        return self._started_at

    def has_modifiers(self):
        return bool(self._modifiers)

    def modifiers(self):
        return self._modifiers

    def end(self):
        self._stopped_at = datetime.utcnow()

    def is_running(self):
        return self._stopped_at is None

    def ended_at(self):
        return self._stopped_at


class Simulation:
    """
        Keeps track of running scenarios.
    """

    def __init__(self):
        self._scenarios = []
        self._started_at = None
        self._stopped_at = None

    def started_at(self):
        return self._started_at

    def stopped_at(self):
        return self._stopped_at

    def start(self):
        if self.is_running():
            raise Exception('Simulation is already running')

        self._started_at = datetime.utcnow()

    def start_scenario(self, name: str, modifiers: Optional[dict] = None, options: Optional[dict] = None):
        if not self.is_running():
            raise Exception('Can not start scenario when simulation is not running')

        if self.scenario_is_running():
            self.running_scenario().end()

        self._scenarios.append(RunningScenario(
            name=name,
            started_at=datetime.utcnow(),
            modifiers=modifiers or {},
            options=options
        ))

    def current_modifiers(self):
        if not self.scenario_is_running():
            return {}

        return self.running_scenario().modifiers()

    def stop_running_scenario(self):
        assert self.scenario_is_running(), "No scenario is running"

        self.running_scenario().end()

    def stop(self):
        if not self.is_running():
            raise Exception('Simulation is not running')

        if self.scenario_is_running():
            self.running_scenario().end()

        self._stopped_at = datetime.utcnow()

    def _finalise_last_scenario(self):
        pass

    def scenario_is_running(self):
        return len(self.scenarios()) > 0 and self.scenarios()[-1].is_running()

    def is_running(self) -> bool:
        return self._started_at is not None

    def running_scenario(self) -> Optional[RunningScenario]:
        if len(self.scenarios()) == 0:
            return None

        return self.scenarios()[-1]

    def scenarios(self):
        return self._scenarios


simulation: Simulation = Simulation()



from datetime import timedelta

from flask_scenario_testing.analysis.segmenters.Segmenter import Segmenter
from flask_scenario_testing.analysis.support.Measurement import Measurement
from flask_scenario_testing.analysis.support.MeasurementsSegment import MeasurementsSegment


class TimeWindowedSegmenter(Segmenter):

    def __init__(self, window_size: timedelta):
        self.window_size = window_size

    def segment(self, measurements: [Measurement]) -> [MeasurementsSegment]:
        if len(measurements) == 0:
            return []
        
        window_from = measurements[0].time

        idx = 0
        segments = []

        while window_from <= measurements[-1].time:
            measurements_group = []

            while idx < len(measurements) and measurements[idx].time < window_from + self.window_size:
                measurements_group.append(measurements[idx])

                idx += 1

            segment = MeasurementsSegment(
                start=window_from,
                time=window_from + self.window_size / 2,
                end=window_from + self.window_size,
                measurements=measurements_group
            )
            segments.append(segment)

            window_from = window_from + self.window_size

        return segments



from flask_scenario_testing.analysis.Results import Results
from flask_scenario_testing.analysis.Scenario import Scenario
from flask_scenario_testing.analysis.support.Averager import Averager


class ComputeCpuUsageOverTime(object):
    def __init__(self, results: Results, averager: Averager):
        self.results = results
        self.averager = averager

    def has_averages(self):
        return self.averager is not None

    def cpu_usages_over_time(self):
        cpu_usage_measurements = self.results.cpu_usage_measurements()

        return zip(*cpu_usage_measurements)

    def average_cpu_usages_over_time(self):
        cpu_usage_measurements = self.results.cpu_usage_measurements()

        times, average_cpu_values = zip(*self.averager.average(cpu_usage_measurements))

        return times, average_cpu_values

    def scenarios(self) -> [Scenario]:
        return self.results.scenarios()



import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter

from flask_scenario_testing.analysis.services.ComputeCpuUsageOverTime import ComputeCpuUsageOverTime
from flask_scenario_testing.analysis.report_sections.ReportSection import ReportSection


class PlotCpuUsageOverTime(ReportSection):
    def __init__(self, service: ComputeCpuUsageOverTime):
        ReportSection.__init__(self)
        self.service = service

    def print(self, _) -> None:
        times, cpu_values = self.service.cpu_usages_over_time()

        figure = plt.figure()
        plt.plot(times, cpu_values, label='CPU Usage', color=self.next_color())

        if self.service.has_averages():
            times, cpu_values = self.service.average_cpu_usages_over_time()

            plt.plot(times, cpu_values, label='CPU Usage (average)', color=self.next_color())

        self.draw_scenario_vlines()

        plt.title('CPU Usage over time')
        plt.xticks(rotation=45)

        ax = plt.gca()
        ax.yaxis.set_major_formatter(FuncFormatter(lambda x, _: '{}%'.format(x)))
        ax.set_ylim([0, 110])

        plt.legend()
        plt.gcf().subplots_adjust(bottom=0.2)

        self.add_figure(figure)

    def draw_scenario_vlines(self):
        for scenario in self.service.scenarios():
            label = "Start '{}'".format(scenario.name())

            plt.axvline(x=scenario.started_at(), color=self.next_color(), linestyle='--', label=label)




from flask_scenario_testing.analysis.segmenters.Segmenter import Segmenter
from flask_scenario_testing.analysis.support.Measurement import Measurement


class Averager(object):
    def __init__(self, segmenter: Segmenter, should_round: bool = False):
        self.segmenter = segmenter
        self.should_round = should_round

    def average(self, data):
        segments = self.segmenter.segment(data)

        return [
            Measurement(time=s.time, value=round(s.mean()) if self.should_round else s.mean())
            for s in segments
        ]



import numpy as np

from flask_scenario_testing.analysis.CpuBucket import CpuBucket
from flask_scenario_testing.analysis.Results import Results
from flask_scenario_testing.analysis.Scenario import Scenario

from flask_scenario_testing.analysis.support.Averager import Averager
from flask_scenario_testing.analysis.support.Measurement import Measurement


class ComputeCpuUsageAgainstLatency(object):
    def __init__(self, averager: Averager, results: Results):
        self.results = results
        self.averager = averager

    def get_data(self, scenario: Scenario, endpoint_name: str):
        cpu_usage_groups = self.averager.average(self.results.cpu_usage_measurements())

        latency_measurements = self.results.latency_measurements(
            endpoint_name,
            scenario.started_at(),
            scenario.ended_at()
        )

        bucket = CpuBucket()

        for latency_measurement in latency_measurements:
            cpu_measurement = self.find_nearby_cpu_measurement(cpu_usage_groups, latency_measurement.time)

            bucket.add(cpu_measurement.value, latency_measurement.value)

        cpu_usages, average_latencies = bucket.contents()

        # trend = self.compute_trend(cpu_usages, average_latencies)

        return cpu_usages, average_latencies

    def find_nearby_cpu_measurement(self, cpu_usage_groups, time) -> Measurement:
        diffs = [(idx, abs((c.time - time).total_seconds())) for idx, c in enumerate(cpu_usage_groups)]
        idx, smallest_value = sorted(diffs, key=lambda d: d[1])[0]

        return cpu_usage_groups[idx]

    def compute_trend(self, cpu_usages, average_latencies):
        z = np.polyfit(cpu_usages, average_latencies, 1)
        p = np.poly1d(z)

        return dict(x=cpu_usages, y=p(cpu_usages))



from datetime import datetime, timedelta

from flask_scenario_testing.analysis.segmenters.RollingSegmenter import RollingSegmenter
from flask_scenario_testing.analysis.support.Measurement import Measurement

now = datetime.now()
segmenter = RollingSegmenter(3)


def test_segmenting_window_size_equals_array_size():
    inputs = [
        Measurement(value=9, time=now),
        Measurement(value=12, time=now + timedelta(seconds=5)),
        Measurement(value=15, time=now + timedelta(seconds=10)),
    ]

    segments = segmenter.segment(inputs)
    assert len(segments) == 1

    assert segments[0].time == now + timedelta(seconds=5)

    assert len(segments[0].measurements) == 3

    assert segments[0].measurements[0].value == 9
    assert segments[0].measurements[1].value == 12
    assert segments[0].measurements[2].value == 15


def test_segmenting_window_size_smaller_than_array_size():
    inputs = [
        Measurement(value=9, time=now),
        Measurement(value=12, time=now + timedelta(seconds=5)),
    ]

    results = segmenter.segment(inputs)

    assert len(results) == 0


def test_segmenting_window_size_larger_than_array_size():
    inputs = [
        Measurement(value=100, time=now),
        Measurement(value=200, time=now + timedelta(seconds=5)),
        Measurement(value=300, time=now + timedelta(seconds=10)),
        Measurement(value=400, time=now + timedelta(seconds=15)),
        Measurement(value=500, time=now + timedelta(seconds=20)),
    ]

    results = segmenter.segment(inputs)

    assert len(results) == 3

    # First segment
    assert results[0].measurements[0].value == 100
    assert results[0].measurements[1].value == 200
    assert results[0].measurements[2].value == 300

    # Second segment
    assert results[1].measurements[0].value == 200
    assert results[1].measurements[1].value == 300
    assert results[1].measurements[2].value == 400

    # Third segment
    assert results[2].measurements[0].value == 300
    assert results[2].measurements[1].value == 400
    assert results[2].measurements[2].value == 500



from flask_scenario_testing.backend.modifiers.Modifier import Modifier
import time


def fib(n):
    if n == 1:
        return 0
    elif n == 2:
        return 1
    else:
        return fib(n - 1) + fib(n - 2)


class Fibonacci(Modifier):
    def identifier(self):
        return 'FIB'

    def modify(self, fun, endpoint_name, modifier_args: dict):
        def wrapper(*args, **kwargs):
            N = modifier_args['N']

            start = time.time()
            fib_result = fib(N)
            ms = (time.time() - start) * 1000

            print("FIB({}) = {} ({}ms)".format(N, fib_result, ms))

            return fun(*args, **kwargs)

        return wrapper



from flask_scenario_testing.analysis.support.Measurement import Measurement
import numpy as np

from flask_scenario_testing.support.pick import pick


class MeasurementsSegment(object):
    def __init__(self, time, measurements: [Measurement], start, end):
        self.time = time
        self.measurements = measurements
        self.start = start
        self.end = end

    def mean(self):
        if len(self.measurements) == 0:
            return None

        return np.median(pick(self.measurements, 'value'))



from datetime import timedelta, datetime

from flask_scenario_testing.analysis.segmenters.TimeWindowedSegmenter import TimeWindowedSegmenter
from flask_scenario_testing.analysis.support.Measurement import Measurement

segmenter = TimeWindowedSegmenter(timedelta(seconds=6))
now = datetime.now()


def test_stuff():
    inputs = [
        Measurement(value=9, time=now),
        Measurement(value=12, time=now + timedelta(seconds=4)),
        Measurement(value=15, time=now + timedelta(seconds=8)),
    ]

    segments = segmenter.segment(inputs)

    assert len(segments) == 2

    # Check segment 1
    assert len(segments[0].measurements) == 2
    assert segments[0].measurements[0].value == 9
    assert segments[0].measurements[1].value == 12
    assert segments[0].time == now + timedelta(seconds=3)

    # Check segment 2
    assert len(segments[1].measurements) == 1
    assert segments[1].measurements[0].value == 15
    assert segments[1].time == now + timedelta(seconds=9)


def test_with_measurements_on_the_border_of_windows():
    inputs = [
        Measurement(value=9, time=now),
        Measurement(value=12, time=now + timedelta(seconds=6)),
        Measurement(value=15, time=now + timedelta(seconds=12)),
    ]

    segments = segmenter.segment(inputs)

    # Segment 1
    assert len(segments[0].measurements) == 1
    assert segments[0].measurements[0].value == 9
    assert segments[0].time == now + timedelta(seconds=3)

    # Segment 2
    assert len(segments[1].measurements) == 1
    assert segments[1].measurements[0].value == 12
    assert segments[1].time == now + timedelta(seconds=9)

    # Segment 3
    assert len(segments[2].measurements) == 1
    assert segments[2].measurements[0].value == 15
    assert segments[2].time == now + timedelta(seconds=15)



class Host(object):
    def __init__(self, hostname: str, port: str):
        self.hostname = hostname
        self.port = port

    def url(self, path=None):
        base = 'http://{}:{}'.format(self.hostname, self.port)

        if not path:
            return base

        return '{}/{}'.format(base, path)


from flask_scenario_testing.analysis.segmenters.Segmenter import Segmenter
from flask_scenario_testing.analysis.support.Measurement import Measurement
from flask_scenario_testing.analysis.support.MeasurementsSegment import MeasurementsSegment


class RollingSegmenter(Segmenter):
    """
        Segments measurements by using a rolling window. Example for N=3:

        [1,2,3,4,5,6,7] => [1,2,3], [2,3,4], [4,5,6], [5,6,7]
    """

    def __init__(self, n, step=1):
        self.n = n
        self.step = step

    def segment(self, measurements: [Measurement]) -> [MeasurementsSegment]:
        center = int(self.n / 2)

        segments = []

        while center + (self.n - int(self.n / 2)) <= len(measurements):
            time = measurements[center].time

            relevant_measurements = measurements[center - int(self.n / 2): center + int(self.n / 2) + 1]

            segments.append(MeasurementsSegment(
                time=time,
                measurements=relevant_measurements,
                start=relevant_measurements[0].time,
                end=relevant_measurements[-1].time
            ))

            center += self.step

        return segments



class Time(object):

    def __init__(self, seconds):
        self._seconds = seconds

    def seconds(self) -> int:
        return self._seconds

    @staticmethod
    def from_string(s: str):
        num = int(s[0:-1])

        if s.endswith('s'):
            return Time(num)
        elif s.endswith('m'):
            return Time(num * 60)

    def __str__(self):
        return "{}s".format(self._seconds)



import abc


class Modifier(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def identifier(self):
        pass

    @abc.abstractmethod
    def modify(self, fun, endpoint_name, args):
        pass



import datetime

from flask_scenario_testing.simulation.Api import Api
from flask_scenario_testing.simulation.ScenariosConfig import ScenariosConfig
import subprocess
import json
from time import sleep

from flask_scenario_testing.support.Time import Time


class Simulation(object):
    def __init__(self, api: Api, config: ScenariosConfig, locustfile: str):
        self.locustfile = locustfile
        self.api = api
        self.config = config

    def run(self, outputfile: str):
        print('Starting simulation')
        self.api.start_simulation()
        print('Successfully started simulation')

        scenarios = self.config.scenarios()
        print('Total number of scenarios: {}'.format(len(scenarios)))

        for i, scenario in enumerate(scenarios):
            print('Deploying scenario "{}"'.format(scenario.name()))

            self.api.start_scenario(scenario.name(), scenario.modifiers(), scenario.users_count(), scenario.options())

            print('Starting load ...')
            self._run_locust(scenario.users_count(), scenario.run_time())

            self.api.stop_running_scenario()

            is_last_scenario = i == len(scenarios) - 1

            if not is_last_scenario and self.config.cooldown_time().seconds() > 0:
                self.await_cooldown(self.config.cooldown_time())

            print('Done')

        print('Stopping simulation')
        result = self.api.stop_simulation()
        print('Successfully stopped simulation, writing results to disk: {}'.format(outputfile.split('/')[-1]))

        with open(outputfile, 'w') as outfile:
            json.dump(result, outfile)

    def _run_locust(self, users_count, total_run_time: Time):
        args = [
            'locust',
            '-f', self.locustfile,
            '--host', self.api.host_url(),
            '--headless',
            '-u', str(users_count),
            '-r', str(users_count),
            '-t', str(total_run_time),
        ]
        p = subprocess.Popen(args)

        run_time_left = total_run_time.seconds()

        while p.poll() is None:
            progress = round((total_run_time.seconds() - run_time_left) / total_run_time.seconds() * 100, 1)
            print('\rSimulating scenario ... {}%'.format(progress), end='')
            sleep(1)
            run_time_left -= 1

        print()

    def await_cooldown(self, cooldown: Time):
        run_time_left = cooldown.seconds()

        while run_time_left > 0:
            progress = round((cooldown.seconds() - run_time_left) / cooldown.seconds() * 100, 1)

            print('\rCooling down ... {}%'.format(progress), end='')
            sleep(1)
            run_time_left -= 1

        print()



import requests

from flask_scenario_testing.simulation.Host import Host


class Api(object):
    def __init__(self, host: Host):
        self.host = host

    def start_simulation(self):
        response = requests.post(self.host.url('api/simulation/start'))

        assert response.status_code == 200, 'Request to start simulation failed, are you sure the host is up? (status code: {})'.format(
            response.status_code)

    def stop_simulation(self):
        response = requests.post(self.host.url('api/simulation/stop'))

        assert response.status_code == 200, 'Request to stop simulation failed due to an unknown error'

        return response.json()

    def start_scenario(self, name, modifiers, meta, options):
        response = requests.post(self.host.url('/api/simulation/start-scenario'), json=dict(
            modifiers=modifiers,
            meta=meta,
            name=name,
            options=options
        ))

        assert response.status_code == 200

    def host_url(self):
        return self.host.url()

    def stop_running_scenario(self):
        response = requests.post(self.host.url('/api/simulation/stop-running-scenario'))

        assert response.status_code == 200

    def measure_travel_time(self):
        response = requests.post(self.host.url('/api/measure-travel-time'))

        assert response.status_code == 200

        return response.json()



class Scenario(object):

    def __init__(self, name, started_at, ended_at):
        self._name = name
        self._started_at = started_at
        self._ended_at = ended_at

    def name(self):
        return self._name

    def started_at(self):
        return self._started_at

    def ended_at(self):
        return self._ended_at



from flask_scenario_testing.analysis.support import MeasurementsSegment, Measurement


class Segmenter(object):
    def segment(self, measurements: [Measurement]) -> [MeasurementsSegment]:
        pass



import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter

from flask_scenario_testing.analysis.report_sections.ReportSection import ReportSection
from flask_scenario_testing.analysis.services.ComputeTransactionSignature import ComputeTransactionSignature
import json

class PlotTransactionSignatures(ReportSection):
    def __init__(self, service: ComputeTransactionSignature, options):
        ReportSection.__init__(self)
        self.color_idx = 0
        self.service = service

    def print(self, _) -> None:
        signatures = [(scenario, self.service.compute_signature(scenario)) for scenario in self.service.scenarios()]

        result = []

        for idx, service_time in enumerate(signatures[0][1]):
            result.append(dict(
                name=service_time.endpoint_name,
                old_service_time=service_time.value,
                new_service_time=signatures[1][1][idx].value
            ))


        print(json.dumps(result, indent=2))
        print(result)
        exit(-1)

        # self.plot_absolute_signatures(signatures)
        # self.plot_relative_signatures(signatures, show_percents=False)
        self.plot_recall_and_precision(signatures)
        # self.plot_relative_signatures(signatures, show_percents=True)

    def plot_absolute_signatures(self, signatures):
        figure = plt.figure()
        ax = plt.gca()
        ax.grid(True, which='both')

        plt.title('Transaction signatures (absolute)')

        for scenario, signature in signatures:
            plt.gcf().subplots_adjust(bottom=0.5)
            plt.xticks(rotation=80)
            plt.plot([s.endpoint_name for s in signature], [s.value for s in signature], label=scenario.name())

        plt.xlabel('Endpoint')
        plt.legend()
        self.add_figure(figure)

    def plot_relative_signatures(self, signatures, show_percents=False):
        figure = plt.figure()
        ax = plt.gca()
        ax.grid(True, which='both')

        base_scenario, base_signature = signatures[0]

        for scenario, signature in signatures[1:-1]:
            plt.gcf().subplots_adjust(bottom=0.5)
            plt.xticks(rotation=80)
            x_values = [s.endpoint_name for s in signature]

            if not show_percents:
                plt.title('Transaction signatures (absolute difference)')

                ax = plt.gca()
                ax.yaxis.set_major_formatter(FuncFormatter(lambda x, _: '{}ms'.format(x)))
                y_values = [s.value - base_signature[idx].value for idx, s in enumerate(signature)]
            else:
                plt.title('Transaction signatures (percentual difference)')

                ax = plt.gca()
                ax.yaxis.set_major_formatter(FuncFormatter(lambda x, _: '{}%'.format(x)))
                y_values = [(s.value - base_signature[idx].value) / base_signature[idx].value * 100 for idx, s in
                            enumerate(signature)]

            plt.plot(x_values, y_values, label=scenario.name())

        plt.xlabel('Endpoint')
        plt.legend()
        self.add_figure(figure)

    def plot_recall_and_precision(self, signatures):
        figure = plt.figure()
        signature1 = signatures[0][1]
        signature2 = signatures[1][1]

        diffs = []
        for i, service_time_1 in enumerate(signature1):
            diff = service_time_1.value - signature2[i].value
            diffs.append(diff)

        start = min(diffs) - 10
        end = max(diffs) + 10

        x_values = []
        false_positive_counts = []
        while start < end:
            x_values.append(start)
            false_positive_count = 0
            for i, service_time_1 in enumerate(signature1):
                diff = service_time_1.value - signature2[i].value

                if diff > start and service_time_1.endpoint_name != 'api.add_user':
                    false_positive_count += 1

            false_positive_counts.append(false_positive_count)
            start += 1

        plt.plot(x_values, false_positive_counts)
        self.add_figure(figure)



from flask_scenario_testing.backend.modifiers.Modifier import Modifier
from time import sleep


class Sleep(Modifier):
    def identifier(self):
        return 'SLEEP'

    def modify(self, fun, endpoint_name, modifier_args: dict):
        def wrapper(*args, **kwargs):
            sleep_for = int(modifier_args.get('time')) / 1000
            sleep(sleep_for)

            print('Sleeping for {}ms'.format(sleep_for))

            return fun(*args, **kwargs)

        return wrapper


from flask_scenario_testing.analysis.Results import Results
import matplotlib.colors as mcolors


class ReportSection(object):

    def __init__(self):
        self.color_idx = 0

    def next_color(self) -> str:
        colors = list(mcolors.TABLEAU_COLORS.items())

        c = colors[self.color_idx][0]

        self.color_idx += 1

        return c

    def print(self, results: Results):
        pass

    def print_row(self, cols, col_widths: []):
        for idx, col in enumerate(cols):
            print(str(col).ljust(col_widths[idx]), end='')

        print()

    def print_separator(self, length=40):
        print('-' * length)

    def table(self, headers, data):
        print('| ' + ' | '.join(headers) + ' | ')
        print('| ' + ' | '.join(['-' * max(1, len(header)) for header in headers]) + ' | ')

        for row in data:
            print('| ' + ' | '.join([str(item) for item in row]) + ' | ')

        print()

    def heading(self, text, level=1):
        print('#' * level + ' ' + text)

    def add_figure(self, figure):
        figure.show()



from flask_scenario_testing.simulation.ScenarioConfig import ScenarioConfig
import json

from flask_scenario_testing.support.Time import Time


class ScenariosConfig(object):
    def __init__(self, scenarios: [ScenarioConfig], defaults: dict, cooldown: Time):
        self._scenarios = scenarios
        self._defaults = defaults
        self._cooldown = cooldown

    def scenarios(self) -> [ScenarioConfig]:
        return self._scenarios

    def defaults(self):
        return self._defaults

    def cooldown_time(self) -> Time:
        return self._cooldown

    @staticmethod
    def from_json(file_path: str):
        scenarios = []

        with open(file_path, 'r') as f:
            configuration = json.load(f)
            defaults = configuration.get('defaults', {})

            if configuration.get('cooldown'):
                cooldown = Time.from_string(configuration.get('cooldown'))
            else:
                cooldown = Time(0)

            for idx, dict_scenario in enumerate(configuration['scenarios']):
                scenarios.append(ScenarioConfig(
                    dict_scenario.get('name', "Scenario '{}'".format(idx + 1)),
                    dict_scenario.get('users', defaults.get('users')),
                    dict_scenario.get('duration', defaults.get('duration')),
                    dict_scenario.get('modifiers', defaults.get('modifiers', [])),
                    dict_scenario.get('options', defaults.get('options', dict()))
                ))

        return ScenariosConfig(scenarios, defaults, cooldown)



import numpy as np


class CpuBucket(object):
    def __init__(self):
        self.bucket_collection = {}
        self.counts = {}

        for i in range(0, 101):
            self.bucket_collection[i] = []
            self.counts[i] = 0

    def add(self, cpu_value: int, latency_measurement: float):
        assert 0 <= cpu_value <= 100

        self.bucket_collection[cpu_value].append(latency_measurement)
        self.counts[cpu_value] += 1

    def contents(self):
        averages_arr = []
        cpu_usages = []

        for i in range(0, 101):
            bucket = self.bucket_collection[i]

            if len(bucket) > 0:
                cpu_usages.append(i)
                averages_arr.append(np.mean(self.bucket_collection[i]))

        return cpu_usages, averages_arr



from matplotlib.ticker import FuncFormatter

from flask_scenario_testing.analysis.Results import Results
import matplotlib.pyplot as plt

from flask_scenario_testing.analysis.services.ComputeCpuUsageAgainstLatency import ComputeCpuUsageAgainstLatency
from flask_scenario_testing.analysis.report_sections.ReportSection import ReportSection


class PlotCpuUsageAgainstLatency(ReportSection):

    def __init__(self, endpoint_name: str, service: ComputeCpuUsageAgainstLatency):
        self.service = service
        self.endpoint_name = endpoint_name

    def print(self, results: Results):
        figure = plt.figure()

        ax = plt.gca()
        ax.xaxis.set_major_formatter(FuncFormatter(lambda x, _: '{}%'.format(x)))
        ax.yaxis.set_major_formatter(FuncFormatter(lambda x, _: '{} ms'.format(x)))

        for scenario in results.scenarios():
            cpu_usages, latencies = self.service.get_data(scenario, self.endpoint_name)
            plt.plot(cpu_usages, latencies, '--*', label=scenario.name())

            # plt.plot(trend['x'], trend['y'], 'r--', label='Trend line')

        plt.ylabel('Latency')
        plt.xlabel('CPU Usage')
        plt.legend()

        # max_y_value = max(latencies)

        # Set CPU usage limit to [0, 100]
        ax.set_xlim([0, 100])

        plt.title('CPU Usage vs average latency ({})'.format(self.endpoint_name))

        self.add_figure(figure)



from flask_scenario_testing.analysis.Results import Results
from flask_scenario_testing.analysis.report_sections.ReportSection import ReportSection
import matplotlib.pyplot as plt

from flask_scenario_testing.analysis.support.Averager import Averager
from flask_scenario_testing.support.pick import pick


class PlotLatenciesOverTime(ReportSection):
    def __init__(self, endpoint_names: [str], averager: Averager):
        ReportSection.__init__(self)
        self.endpoint_names = endpoint_names
        self.averager = averager

    def draw_scenario_vlines(self, results: Results):
        for scenario in results.scenarios():
            label = "Start '{}'".format(scenario.name())

            plt.axvline(x=scenario.started_at(), color=self.next_color(), linestyle='--', label=label)

    def print(self, results: Results):
        figure = plt.figure()

        for endpoint_name in self.endpoint_names:
            latency_measurements = results.latency_measurements(endpoint_name)

            avg_measurements = self.averager.average(latency_measurements)

            y_values = pick(avg_measurements, 'value')
            x_values = pick(avg_measurements, 'time')

            plt.ylabel('Latency')
            plt.xlabel('Time')

            plt.plot(x_values, y_values, label=endpoint_name)

        plt.title('Latency over time')

        self.draw_scenario_vlines(results)
        plt.legend()

        self.add_figure(figure)



from flask_scenario_testing.support.Time import Time


class ScenarioConfig(object):
    def __init__(self, name, users_count: int, run_time: str, modifiers, options):
        self._users_count = users_count
        self._run_time = run_time
        self._modifiers = modifiers
        self._name = name
        self._options = options

    def users_count(self):
        return self._users_count

    def run_time(self) -> Time:
        return Time.from_string(self._run_time)

    def modifiers(self):
        return self._modifiers

    def name(self):
        return self._name

    def options(self):
        return self._options



from flask_scenario_testing.support.Time import Time


def test_creating_seconds():
    time = Time.from_string('60s')

    assert time.seconds() == 60


def test_creating_minutes():
    time = Time.from_string('30m')

    assert time.seconds() == 1800


def test_converting_to_string():
    time = Time.from_string('30m')

    assert (str(time)) == '1800s'



import argparse
import os
from datetime import timedelta

from matplotlib import pyplot as plt

from flask_scenario_testing.analysis.Results import Results
from flask_scenario_testing.analysis.report_sections.PlotTransactionSignatures import PlotTransactionSignatures
from flask_scenario_testing.analysis.segmenters.TimeWindowedSegmenter import TimeWindowedSegmenter
from flask_scenario_testing.analysis.services.ComputeTransactionSignature import ComputeTransactionSignature
from flask_scenario_testing.analysis.support.Averager import Averager
import psutil
import json


def main2():
    with open(os.path.abspath('data.json'), 'r') as f:
        runs = json.load(f)

        plt.figure()

        for run in runs:
            plt.figure()

            print('Processing {}'.format(run['name']))
            print(run['data'])

            diffs = []
            for item in run['data']:
                diffs.append(item['new_service_time'] - item['old_service_time'])

            threshold = min(diffs) - 10
            end = max(diffs) + 10
            all_false_positive_counts = []
            all_true_positive_counts = []
            all_false_negative_counts = []
            all_true_negative_counts = []

            thresholds = []

            while threshold < end:
                thresholds.append(threshold)

                false_positive_count = 0
                true_positive_count = 0
                true_negative_count = 0
                false_negative_count = 0

                for item in run['data']:
                    diff = item['new_service_time'] - item['old_service_time']
                    if diff >= threshold and not item['regression']:
                        false_positive_count += 1
                    if diff >= threshold and item['regression']:
                        true_positive_count += 1
                    if diff < threshold and not item['regression']:
                        true_negative_count += 1
                    if diff < threshold and item['regression']:
                        false_negative_count += 1

                all_false_positive_counts.append(false_positive_count)
                all_true_positive_counts.append(true_positive_count)
                all_false_negative_counts.append(false_negative_count)
                all_true_negative_counts.append(true_negative_count)
                threshold += 1

            plt.plot(thresholds, all_false_positive_counts, label='False positives')
            plt.plot(thresholds, all_true_positive_counts, label='True positives')
            plt.plot(thresholds, all_false_negative_counts, label='False negatives')
            plt.plot(thresholds, all_true_negative_counts, label='True negatives')

            plt.xlabel('Service time diff threshold')
            plt.legend()
            plt.show()


        print(runs[0])
        exit(-1)


def main():
    parser = argparse.ArgumentParser(description='Run a simulation')
    parser.add_argument('file', metavar='file', type=str, help='File to analyse')
    args = parser.parse_args()

    absolute_output_dir = os.path.abspath(args.file)

    psutil.cpu_percent()
    results = Results.from_json(absolute_output_dir)

    segmenter = TimeWindowedSegmenter(timedelta(seconds=30))
    averager = Averager(segmenter, should_round=False)
    rounding_averager = Averager(segmenter, should_round=True)

    summary_sections = [
        # EndpointOverview(),
        # SimulationDetails(),
        # PlotCpuUsageOverTime(service=ComputeCpuUsageOverTime(results, averager=averager)),
        # PlotCpuUsageAgainstLatency('api.add_user', service=ComputeCpuUsageAgainstLatency(
        #     results=results,
        #     averager=rounding_averager)
        # ),
        # PlotLatenciesOverTime(['articles.get_articles', 'articles.favorite_an_article'], averager),
        PlotTransactionSignatures(
            ComputeTransactionSignature(results, segmenter), options=dict(
                plot_absolute=True,
                plot_relative=False,
                plot_percentual_relative=False
            )
        )
    ]

    for section in summary_sections:
        section.print(results)
        print()


if __name__ == '__main__':
    main2()



import pytest

from flask_scenario_testing.backend.Simulation import Simulation


def test_can_start_simulation():
    simulation = Simulation()

    simulation.start()

    assert simulation.is_running()


def test_cant_stop_simulation_thats_not_running():
    simulation = Simulation()

    with pytest.raises(Exception):
        simulation.stop()


def test_can_not_start_simulation_thats_already_running():
    simulation = Simulation()
    simulation.start()

    with pytest.raises(Exception):
        simulation.start()


def test_can_start_scenario():
    simulation = Simulation()
    simulation.start()

    simulation.start_scenario('5 users')

    assert simulation.scenario_is_running()

    running_scenario = simulation.running_scenario()
    assert running_scenario.name() == '5 users'
    assert running_scenario.started_at()
    assert not running_scenario.has_modifiers()


def test_starting_a_second_scenario_will_stop_previous_scenario():
    simulation = Simulation()
    simulation.start()

    simulation.start_scenario('5 users')
    assert simulation.scenarios()[0].is_running()

    simulation.start_scenario('10 users')
    assert not simulation.scenarios()[0].is_running()
    assert simulation.scenarios()[1].is_running()

    simulation.start_scenario('15 users')
    assert not simulation.scenarios()[0].is_running()
    assert not simulation.scenarios()[1].is_running()
    assert simulation.scenarios()[2].is_running()


def test_can_stop_running_scenario():
    simulation = Simulation()

    simulation.start()
    simulation.start_scenario('5 users')
    simulation.stop_running_scenario()

    assert not simulation.scenario_is_running()
    assert simulation.scenarios()[0].started_at() is not None
    assert not simulation.scenarios()[0].is_running()


def test_can_convert_to_dict():
    simulation = Simulation()

    simulation.start()
    simulation.start_scenario('5 users')


from collections import namedtuple

from flask_scenario_testing.support.pick import pick


def test_pick():
    Point = namedtuple('Point', ['x', 'y'])

    points = [Point(x=3, y=5), Point(x=4, y=0)]

    assert pick(points, 'x')[0] == 3
    assert pick(points, 'x')[1] == 4

    assert pick(points, 'y')[0] == 5
    assert pick(points, 'y')[1] == 0



import time

from flask_scenario_testing.backend.modifiers.Modifier import Modifier


class BusySleep(Modifier):
    def identifier(self):
        return 'BUSY_SLEEP'

    def modify(self, fun, endpoint_name, modifier_args: dict):
        def wrapper(*args, **kwargs):
            sleep_for = int(modifier_args.get('time')) / 1000

            time_start = time.time()

            while time.time() < time_start + sleep_for:
                pass

            return fun(*args, **kwargs)

        return wrapper



import datetime

from locust import task, HttpUser, between, SequentialTaskSet
import random
import time
from urllib.parse import urlencode


class LoadTest(SequentialTaskSet):
    _base_headers = {'Content-Type': 'application/json'}
    limit = 10
    offset = 0
    current_article = None
    query_params = {}
    session_id = None

    _recommended_articles = []
    _interesting_topics = []
    _subscribed_topics = []

    def headers(self):
        return {
            'sent-at': str(time.time())
        }

    def make_url(self, url, query_params: dict = None):
        if not query_params:
            query_params = {}

        if self.session_id:
            query_params['session'] = self.session_id

        encoded_query_string = urlencode(query_params)

        if encoded_query_string == '':
            return url

        return '{}?{}'.format(url, urlencode(query_params))

    def post(self, url, query_params: dict = None, data=None):
        url = self.make_url(url, query_params)

        return self.client.post(
            url,
            headers=self.headers(),
            data=data
        )

    def get(self, url, query_params: dict = None):
        if not url.startswith('/'):
            url = '/' + url

        url = self.make_url(url, query_params)

        return self.client.get(url)

    def subscribe_to_random_topic(self):
        interesting_topic = random.choice(self._interesting_topics)
        print('Subscribing to topic {}'.format(interesting_topic['title']))
        self.post('/subscribe_topic', data=dict(
            topic_id=interesting_topic['id']
        ))

    def navigate_to_home(self):
        self._recommended_articles = self.get('/user_articles/recommended').json()
        self._interesting_topics = self.get('/interesting_topics').json()
        self._subscribed_topics = self.get('/subscribed_topics').json()

    def read_random_article(self):
        """
        Fetches an article, uploads some user activity
        """
        article = random.choice(self._recommended_articles)
        print("Reading article {}".format(article['title']))
        self.get('user_article', query_params=dict(
            article_id=article['id']
        ))
        time.sleep(1)
        self.upload_scroll_activity()
        time.sleep(1)
        self.upload_scroll_activity()

    def upload_scroll_activity(self):
        response = self.post('/upload_user_activity_data', data=dict(
            event='UMR - SCROLL',
            extra_data='',
            time=datetime.datetime.utcnow().isoformat() + 'Z'
        ))

        if response.status_code != 200:
            print({'content':   response.content, 'status_code': response.status_code})
            exit(-1)

    @task()
    def task_register(self):
        """
        Register leads to the home page, which makes the articles load, as well as the interesting topics and the
        subscribed topics
        """
        random_number = random.randint(0, 10000000)

        response = self.post('/add_user/johan+{}@gmail.com'.format(random_number),
                         data=dict(password='password123', invite_code='please'))
        assert response.status_code == 200
        self.session_id = response.content.decode('utf-8')



        # Registering redirects to /articles page
        self.navigate_to_home()

    @task()
    def task_subscribe_to_topic(self):
        self.subscribe_to_random_topic()

    @task()
    def task_subscribe_to_another_topic(self):
        self.subscribe_to_random_topic()

    @task()
    def task_read_article(self):
        self.read_random_article()

    @task()
    def task_go_back_home(self):
        self.navigate_to_home()

    @task()
    def task_read_another_article(self):
        self.read_random_article()

    @task()
    def task_view_settings_page(self):
        self.get('/get_user_details')
        self.get('/system_languages')

    @task()
    def task_update_settings_page(self):
        self.post('/user_settings', data=dict(
            learned_language='fr',
            native_language='nl'
        ))
        self.navigate_to_home()


class User(HttpUser):
    tasks = [LoadTest]

    wait_time = between(7, 15)



import numpy as np

from flask_scenario_testing.analysis.Results import Results

from flask_scenario_testing.analysis.report_sections.ReportSection import ReportSection
from flask_scenario_testing.support.pick import pick


class MedianLatency(ReportSection):
    def print(self, results: Results):
        endpoints = results.endpoints()

        data = []
        for endpoint in endpoints:
            measurements = pick(endpoint['latency_measurements'], 'measurement')

            data.append([
                endpoint['name'],
                np.median(measurements) if len(measurements) > 0 else 'N/A'
            ])

        self.table(['Endpoint', 'Measurement'], data)



from flask_scenario_testing.backend.modifiers.Modifier import Modifier
from time import sleep
import random


class Sleep(Modifier):
    def identifier(self):
        return 'RANDOM_SLEEP'

    def modify(self, fun, endpoint_name, modifier_args: dict):
        def wrapper(*args, **kwargs):
            base_sleep_ms = int(modifier_args.get('time'))
            random_offset_ms = random.randint(-200, +200)

            sleep_ms = base_sleep_ms + random_offset_ms

            print('Sleeping for {}ms + {}ms = {}ms'.format(base_sleep_ms, random_offset_ms, sleep_ms))
            sleep(sleep_ms / 1000)

            return fun(*args, **kwargs)

        return wrapper



import setuptools
import os


loc = os.path.dirname(os.path.abspath(__file__))

with open(loc + '/requirements.txt') as f:
    required = f.read().splitlines()


setuptools.setup(
    name="flask-scenario-testing",
    scripts=['bin/run-simulation'],
    version='1.0.15',
    setup_requires=['setuptools_scm'],
    packages=setuptools.find_packages(),
    include_package_data=True,
    platforms='Any',
    zip_safe=False,
    test_suite='tests.get_test_suite',
    url='https://github.com/thejager/flask-scenario-testing',
    author="Johan de Jager, Mircea Lungu",
    author_email="johanthejager@gmail.com",
    description="Automatically monitor the evolving performance of Flask/Python web services.",
    install_requires=required,
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Framework :: Flask',
    ],
    project_urls={
    },
)



# coding: utf-8
from collections import namedtuple
from datetime import datetime

from flask import Blueprint, jsonify, request
from flask_scenario_testing.backend.Simulation import simulation

blueprint = Blueprint('scenario-testing', __name__)

Measurement = namedtuple('Measurement', ['time', 'measurement'])


def success():
    return jsonify({'status': 'success'})


def get_cpu_usage_measurements(start: datetime, end: datetime):
    from flask_scenario_testing.backend.initialise_scenario_testing import cpu_percents_measurements

    return [
        {'time': measurement.time.isoformat(), 'measurement': measurement.value}
        for measurement in cpu_percents_measurements
        if start <= measurement.time <= end
    ]


@blueprint.route('/api/simulation/start', methods=('POST',))
def start_simulation():
    print('Starting simulation')

    simulation.start()

    return success()


@blueprint.route('/api/simulation/start-scenario', methods=('POST',))
def start_new_scenario():
    data = request.json

    print('Starting new scenario')

    simulation.start_scenario(data['name'], data['modifiers'], data['options'])

    return success()


@blueprint.route('/api/simulation/stop-running-scenario', methods=('POST',))
def stop_running_scenario():
    print('Stopping running scenario')

    simulation.stop_running_scenario()

    return success()


@blueprint.route('/api/simulation/stop', methods=('POST',))
def stop_simulation():
    from flask_scenario_testing.backend.initialise_scenario_testing import latency_measurements

    simulation.stop()

    endpoints = []
    for (endpoint_name, measurements) in latency_measurements.items():
        endpoints.append(dict(
            active=len(measurements) > 0,
            latency_measurements=[dict(
                measurement=m.value,
                time=m.time.isoformat()
            ) for m in measurements],
            name=endpoint_name,
            request_count=len(measurements)
        ))

    cpu_usage_measurements = get_cpu_usage_measurements(simulation.started_at(), simulation.stopped_at())

    scenario_data = []

    for scenario in simulation.scenarios():
        scenario_data.append(dict(
            start=scenario.started_at().isoformat(),
            end=scenario.ended_at().isoformat(),
            modifiers=scenario.modifiers(),
            name=scenario.name()
        ))

    return jsonify(dict(
        started_at=simulation.started_at().isoformat(),
        stopped_at=simulation.stopped_at().isoformat(),
        scenarios=scenario_data,
        endpoints=endpoints,
        cpu_usage_measurements=cpu_usage_measurements
    ))


@blueprint.route('/api/measure-travel-time', methods=('POST',))
def measure_travel_time():
    return jsonify(dict(
        current_time=datetime.utcnow().isoformat()
    ))



from datetime import datetime
from functools import wraps
import time

from apscheduler.schedulers.background import BackgroundScheduler
from flask import Flask, request
import flask_scenario_testing.backend.views as scenario_testing_views
import psutil

from flask_scenario_testing.analysis.support.Measurement import Measurement
from flask_scenario_testing.backend.Simulation import simulation
from flask_scenario_testing.backend.modifiers import Modifier

registered_modifier_types = {}
scheduler = BackgroundScheduler()
cpu_percents_measurements = []
latency_measurements = {}


def add_modifier_wrapper(app, endpoint_name):
    f = app.view_functions[endpoint_name]

    @wraps(f)
    def wrapper(*args, **kwargs):
        relevant_modifiers = [c for c in simulation.current_modifiers() if c['endpoint_name'] == endpoint_name]

        new_f = f
        for modifier in relevant_modifiers:
            regression_type = modifier['type']

            apply = [c for c in registered_modifier_types if c.identifier() == regression_type]

            for a in apply:
                new_f = a.modify(f, endpoint_name, modifier['args'])

        return new_f(*args, **kwargs)

    wrapper.original = f
    app.view_functions[endpoint_name] = wrapper


def add_latency_tracking_wrapper(app, endpoint_name):
    f = app.view_functions[endpoint_name]

    latency_measurements[endpoint_name] = []

    @wraps(f)
    def wrapper(*args, **kwargs):
        start_time = float(request.headers.get('sent-at', time.time()))

        results = f(*args, **kwargs)

        duration_in_ms = (time.time() - start_time) * 1000

        print('{}: {} ms'.format(endpoint_name, duration_in_ms))

        latency_measurements[endpoint_name].append(Measurement(
            time=datetime.utcnow(),
            value=duration_in_ms
        ))

        return results

    wrapper.original = f
    app.view_functions[endpoint_name] = wrapper


def get_cpu_percent():
    return max(psutil.cpu_percent(percpu=True))


def track_cpu_percent():
    cpu_percent = get_cpu_percent()

    cpu_percents_measurements.append(Measurement(time=datetime.utcnow(), value=cpu_percent))

    print('CPU: {}%'.format(cpu_percent))


def start_cpu_usage_tracker():
    scheduler.add_job(func=track_cpu_percent, trigger='interval', seconds=1)
    scheduler.start()


def initialise_scenario_testing(app: Flask, modifier_types: [Modifier]):
    app.register_blueprint(scenario_testing_views.blueprint)

    start_cpu_usage_tracker()

    global registered_modifier_types
    registered_modifier_types = modifier_types

    for rule in app.url_map.iter_rules():
        endpoint_name = rule.endpoint

        if endpoint_name.startswith('scenario-testing'):
            continue

        add_modifier_wrapper(app, endpoint_name)
        add_latency_tracking_wrapper(app, endpoint_name)



import numpy as np

from flask_scenario_testing.analysis.CpuBucket import CpuBucket
from flask_scenario_testing.analysis.Results import Results
from flask_scenario_testing.analysis.segmenters.Segmenter import Segmenter
from flask_scenario_testing.analysis.support.ServiceTime import ServiceTime
from flask_scenario_testing.support.pick import pick


class ComputeTransactionSignature(object):
    def __init__(self, results: Results, segmenter: Segmenter):
        self.results = results
        self.segmenter = segmenter

    def table(self, headers, data):
        print('| ' + ' | '.join(headers) + ' | ')
        print('| ' + ' | '.join(['-' * max(1, len(header)) for header in headers]) + ' | ')

        for row in data:
            print('| ' + ' | '.join([str(item) for item in row]) + ' | ')

        print()

    def compute_signature(self, scenario):
        signature = []

        for endpoint in self.results.endpoints():
            cpu_bucket = CpuBucket()

            latency_measurements = self.results.latency_measurements(
                endpoint['name'],
                scenario.started_at(),
                scenario.ended_at()
            )

            if len(latency_measurements) == 0:
                continue

            segments = self.segmenter.segment(latency_measurements)

            for segment in segments:
                cpu_measurements_in_window = [
                    c.value for c in self.results.cpu_usage_measurements()
                    if segment.start <= c.time < segment.end
                ]

                if len(cpu_measurements_in_window) == 0:
                    print('Skipping segment {} - {}'.format(segment.start, segment.end))
                    continue

                average_cpu_in_monitoring_window = float(np.mean(cpu_measurements_in_window))
                average_rounded_cpu_usage = round(average_cpu_in_monitoring_window)

                latency_measurements = pick(segment.measurements, 'value')

                if len(latency_measurements) == 0:
                    continue

                mean_measurement = float(np.mean(latency_measurements))

                for i in range(len(segment.measurements)):
                    cpu_bucket.add(average_rounded_cpu_usage, mean_measurement)

            cpu_usages, averages_arr = cpu_bucket.contents()

            service_times = []

            print('Endpoint: {}'.format(endpoint['name']))

            data = []
            for cpu_usage, average_latency in zip(cpu_usages, averages_arr):
                if cpu_usage > 80:
                    continue

                service_time = average_latency * (1 - cpu_usage / 100)

                data.append([
                    '{}ms'.format(average_latency),
                    '{}%'.format(cpu_usage),
                    '{}ms'.format(service_time)
                ])

                service_times.append(service_time)

            # self.table(['Latency', 'CPU%', 'Service Time'], data)
            # exit(-1)

            signature.append(ServiceTime(
                endpoint_name=endpoint['name'],
                value=np.median(service_times)
            ))

        return signature

    def scenarios(self):
        return self.results.scenarios()
